"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var CompressedJSON_1 = require("./CompressedJSON");
var Support_1 = require("./Support");
var TypeBuilder_1 = require("./TypeBuilder");
var DateTime_1 = require("./DateTime");
var Type_1 = require("./Type");
function forEachArrayInNestedValueArray(va, f) {
    if (va.length === 0) {
        return;
    }
    if (Array.isArray(va[0])) {
        for (var _i = 0, va_1 = va; _i < va_1.length; _i++) {
            var x = va_1[_i];
            forEachArrayInNestedValueArray(x, f);
        }
    }
    else {
        f(va);
    }
}
function forEachValueInNestedValueArray(va, f) {
    forEachArrayInNestedValueArray(va, function (a) { return a.forEach(f); });
}
var InferenceUnionBuilder = /** @class */ (function (_super) {
    __extends(InferenceUnionBuilder, _super);
    function InferenceUnionBuilder(typeBuilder, _typeInference, _cjson, _fixed) {
        var _this = _super.call(this, typeBuilder, true) || this;
        _this._typeInference = _typeInference;
        _this._cjson = _cjson;
        _this._fixed = _fixed;
        _this.setNumValues = function (n) {
            if (_this._numValues !== undefined) {
                return Support_1.panic("Can only set number of values once");
            }
            _this._numValues = n;
        };
        return _this;
    }
    InferenceUnionBuilder.prototype.makeEnum = function (cases, counts, typeAttributes, forwardingRef) {
        var caseMap = immutable_1.OrderedMap(cases.map(function (c) { return [c, counts[c]]; }));
        return this.typeBuilder.getStringType(typeAttributes, caseMap, forwardingRef);
    };
    InferenceUnionBuilder.prototype.makeClass = function (classes, maps, typeAttributes, forwardingRef) {
        Support_1.assert(maps.length === 0);
        return this._typeInference.inferClassType(this._cjson, typeAttributes, classes, this._fixed, forwardingRef);
    };
    InferenceUnionBuilder.prototype.makeArray = function (arrays, typeAttributes, forwardingRef) {
        return this.typeBuilder.getArrayType(this._typeInference.inferType(this._cjson, typeAttributes, arrays, this._fixed, forwardingRef));
    };
    return InferenceUnionBuilder;
}(TypeBuilder_1.UnionBuilder));
function canBeEnumCase(s) {
    if (s.length === 0)
        return true; // FIXME: Do we really want this?
    return !DateTime_1.isDate(s) && !DateTime_1.isTime(s) && !DateTime_1.isDateTime(s);
}
var TypeInference = /** @class */ (function () {
    function TypeInference(_typeBuilder, _inferEnums, _inferDates) {
        this._typeBuilder = _typeBuilder;
        this._inferEnums = _inferEnums;
        this._inferDates = _inferDates;
    }
    TypeInference.prototype.inferType = function (cjson, typeAttributes, valueArray, fixed, forwardingRef) {
        var _this = this;
        var unionBuilder = new InferenceUnionBuilder(this._typeBuilder, this, cjson, fixed);
        var numValues = 0;
        forEachValueInNestedValueArray(valueArray, function (value) {
            numValues += 1;
            var t = CompressedJSON_1.valueTag(value);
            switch (t) {
                case CompressedJSON_1.Tag.Null:
                    unionBuilder.addNull();
                    break;
                case CompressedJSON_1.Tag.False:
                case CompressedJSON_1.Tag.True:
                    unionBuilder.addBool();
                    break;
                case CompressedJSON_1.Tag.Integer:
                    unionBuilder.addInteger();
                    break;
                case CompressedJSON_1.Tag.Double:
                    unionBuilder.addDouble();
                    break;
                case CompressedJSON_1.Tag.InternedString:
                    if (_this._inferEnums && !unionBuilder.haveString) {
                        var s = cjson.getStringForValue(value);
                        if (canBeEnumCase(s)) {
                            unionBuilder.addEnumCase(s);
                        }
                        else {
                            unionBuilder.addStringType("string");
                        }
                    }
                    else {
                        unionBuilder.addStringType("string");
                    }
                    break;
                case CompressedJSON_1.Tag.UninternedString:
                    unionBuilder.addStringType("string");
                    break;
                case CompressedJSON_1.Tag.Object:
                    unionBuilder.addClass(cjson.getObjectForValue(value));
                    break;
                case CompressedJSON_1.Tag.Array:
                    unionBuilder.addArray(cjson.getArrayForValue(value));
                    break;
                case CompressedJSON_1.Tag.Date:
                    unionBuilder.addStringType(_this._inferDates ? "date" : "string");
                    break;
                case CompressedJSON_1.Tag.Time:
                    unionBuilder.addStringType(_this._inferDates ? "time" : "string");
                    break;
                case CompressedJSON_1.Tag.DateTime:
                    unionBuilder.addStringType(_this._inferDates ? "date-time" : "string");
                    break;
                default:
                    return Support_1.assertNever(t);
            }
        });
        unionBuilder.setNumValues(numValues);
        return unionBuilder.buildUnion(false, typeAttributes, forwardingRef);
    };
    TypeInference.prototype.inferClassType = function (cjson, typeAttributes, objects, fixed, forwardingRef) {
        var propertyNames = [];
        var propertyValues = {};
        forEachArrayInNestedValueArray(objects, function (arr) {
            for (var i = 0; i < arr.length; i += 2) {
                var key = cjson.getStringForValue(arr[i]);
                var value = arr[i + 1];
                if (!Object.prototype.hasOwnProperty.call(propertyValues, key)) {
                    propertyNames.push(key);
                    propertyValues[key] = [];
                }
                propertyValues[key].push(value);
            }
        });
        var properties = [];
        for (var _i = 0, propertyNames_1 = propertyNames; _i < propertyNames_1.length; _i++) {
            var key = propertyNames_1[_i];
            var values = propertyValues[key];
            var t = this.inferType(cjson, immutable_1.Map(), values, false);
            var isOptional = values.length < objects.length;
            properties.push([key, new Type_1.ClassProperty(t, isOptional)]);
        }
        var propertyMap = immutable_1.OrderedMap(properties);
        if (fixed) {
            return this._typeBuilder.getUniqueClassType(typeAttributes, true, propertyMap, forwardingRef);
        }
        else {
            return this._typeBuilder.getClassType(typeAttributes, propertyMap, forwardingRef);
        }
    };
    return TypeInference;
}());
exports.TypeInference = TypeInference;
