"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var TypeBuilder_1 = require("./TypeBuilder");
var Support_1 = require("./Support");
var TypeNames_1 = require("./TypeNames");
var TypeAttributes_1 = require("./TypeAttributes");
function getCliqueProperties(clique, makePropertyType) {
    var properties = immutable_1.OrderedMap();
    for (var _i = 0, clique_1 = clique; _i < clique_1.length; _i++) {
        var c = clique_1[_i];
        c.properties.forEach(function (cp, name) {
            var p = properties.get(name);
            if (p === undefined) {
                p = [immutable_1.OrderedSet(), 0, false];
                properties = properties.set(name, p);
            }
            p[1] += 1;
            p[0] = p[0].union(Type_1.allTypeCases(cp.type));
            if (cp.isOptional) {
                p[2] = true;
            }
        });
    }
    return properties.map(function (_a, name) {
        var types = _a[0], count = _a[1], isOptional = _a[2];
        isOptional = isOptional || count < clique.length;
        var attributes = TypeAttributes_1.combineTypeAttributes(types.map(function (t) { return t.getAttributes(); }).toArray());
        attributes = TypeNames_1.namesTypeAttributeKind.setDefaultInAttributes(attributes, function () { return new TypeNames_1.TypeNames(immutable_1.OrderedSet([name]), immutable_1.OrderedSet(), true); });
        return new Type_1.ClassProperty(makePropertyType(attributes, types), isOptional);
    });
}
var UnifyUnionBuilder = /** @class */ (function (_super) {
    __extends(UnifyUnionBuilder, _super);
    function UnifyUnionBuilder(typeBuilder, _makeEnums, _makeClassesFixed, conflateNumbers, _unifyTypes) {
        var _this = _super.call(this, typeBuilder, conflateNumbers) || this;
        _this._makeEnums = _makeEnums;
        _this._makeClassesFixed = _makeClassesFixed;
        _this._unifyTypes = _unifyTypes;
        return _this;
    }
    UnifyUnionBuilder.prototype.makeEnum = function (enumCases, counts, typeAttributes, forwardingRef) {
        if (this._makeEnums) {
            return this.typeBuilder.getEnumType(typeAttributes, immutable_1.OrderedSet(enumCases), forwardingRef);
        }
        else {
            return this.typeBuilder.getStringType(typeAttributes, immutable_1.OrderedMap(counts), forwardingRef);
        }
    };
    UnifyUnionBuilder.prototype.makeClass = function (classes, maps, typeAttributes, forwardingRef) {
        var _this = this;
        if (maps.length > 0) {
            var propertyTypes_1 = maps.slice();
            for (var _i = 0, classes_1 = classes; _i < classes_1.length; _i++) {
                var classRef = classes_1[_i];
                var c = Type_1.assertIsClass(classRef.deref()[0]);
                c.properties.forEach(function (cp) {
                    propertyTypes_1.push(cp.typeRef);
                });
            }
            var t = this.typeBuilder.getMapType(this._unifyTypes(propertyTypes_1, immutable_1.Map()), forwardingRef);
            this.typeBuilder.addAttributes(t, typeAttributes);
            return t;
        }
        if (classes.length === 1) {
            var t = this.typeBuilder.lookupTypeRef(classes[0]);
            this.typeBuilder.addAttributes(t, typeAttributes);
            return t;
        }
        var maybeTypeRef = this.typeBuilder.lookupTypeRefs(classes);
        // FIXME: Comparing this to `forwardingRef` feels like it will come
        // crashing on our heads eventually.  The reason we need it here is
        // because `unifyTypes` registers the union that we're supposed to
        // build here as a forwarding ref, and we end up with a circular
        // ref if we just return it here.
        if (maybeTypeRef !== undefined && maybeTypeRef !== forwardingRef) {
            this.typeBuilder.addAttributes(maybeTypeRef, typeAttributes);
            return maybeTypeRef;
        }
        var actualClasses = classes.map(function (c) { return Type_1.assertIsClass(c.deref()[0]); });
        var ref;
        ref = this.typeBuilder.getUniqueClassType(typeAttributes, this._makeClassesFixed, undefined, forwardingRef);
        var properties = getCliqueProperties(actualClasses, function (names, types) {
            Support_1.assert(types.size > 0, "Property has no type");
            return _this._unifyTypes(types.map(function (t) { return t.typeRef; }).toArray(), names);
        });
        this.typeBuilder.setClassProperties(ref, properties);
        return ref;
    };
    UnifyUnionBuilder.prototype.makeArray = function (arrays, typeAttributes, forwardingRef) {
        var ref = this.typeBuilder.getArrayType(this._unifyTypes(arrays, immutable_1.Map()), forwardingRef);
        this.typeBuilder.addAttributes(ref, typeAttributes);
        return ref;
    };
    return UnifyUnionBuilder;
}(TypeBuilder_1.UnionBuilder));
function unifyTypes(types, typeAttributes, typeBuilder, makeEnums, makeClassesFixed, conflateNumbers, maybeForwardingRef) {
    if (types.isEmpty()) {
        return Support_1.panic("Cannot unify empty set of types");
    }
    else if (types.count() === 1) {
        var first = Support_1.defined(types.first());
        if (!(first instanceof Type_1.UnionType)) {
            return typeBuilder.lookupTypeRef(first.typeRef);
        }
    }
    var typeRefs = types.toArray().map(function (t) { return t.typeRef; });
    var maybeTypeRef = typeBuilder.lookupTypeRefs(typeRefs);
    if (maybeTypeRef !== undefined) {
        return maybeTypeRef;
    }
    var unionBuilder = new UnifyUnionBuilder(typeBuilder, makeEnums, makeClassesFixed, conflateNumbers, function (trefs, names) {
        return unifyTypes(immutable_1.Set(trefs.map(function (tref) { return tref.deref()[0]; })), names, typeBuilder, makeEnums, makeClassesFixed, conflateNumbers);
    });
    types.forEach(function (t) { return TypeBuilder_1.addTypeToUnionAccumulator(unionBuilder, t); });
    return typeBuilder.withForwardingRef(maybeForwardingRef, function (forwardingRef) {
        typeBuilder.registerUnion(typeRefs, forwardingRef);
        return unionBuilder.buildUnion(false, typeAttributes, forwardingRef);
    });
}
exports.unifyTypes = unifyTypes;
